# Test config for having tools in $PATH - to be installed along with the
# test runners in /usr/share/xdp-tools

XDPDUMP=xdpdump
XDP_BENCH=xdp-bench
XDP_FILTER=xdp-filter
XDP_FORWARD=xdp-forward
XDP_LOADER=xdp-loader
XDP_MONITOR=xdp-monitor
XDP_TRAFFICGEN=xdp-trafficgen
